@@ -3,9 +3,11 @@ from . import tnc, rigctl, freedv, rf
 import platform
 import logging
 from threading import Lock
+import threading
 import random
 import argparse
 import sys
+import time
 
 def main():
     logger = logging.getLogger()
@@ -26,10 +28,13 @@ def main():
     parser.add_argument('--no-tx', action='store_false', help="Disables serial port", dest="tx")
     parser.add_argument('-v', action='store_true', help="Enables debug log", dest="verbose")
     parser.add_argument('--stdout', action='store_true', help="RX data stdout", dest="stdout")
+    parser.add_argument('--tcp', action='store_true', help="Enabled the TCP Server (listens on 0.0.0.0:8001", dest="tcp")
 
     group = parser.add_argument_group(title='Advanced TNC Tuneable', description="These settings can be used to tune the TNC")
-    group.add_argument('--preamble-length', dest="preamble_length", default=10, help="Number of preamble frames to send", type=int)
-    group.add_argument('--min-tx-wait', dest="min_tx_wait", default=5, help="The TNC will wait this value in seconds + a random amount of seconds between 0 and 2", type=int)
+    group.add_argument('--preamble-length', dest="preamble_length", default=9, help="Number of preamble frames to send", type=int)
+    group.add_argument('--min-tx-wait', dest="min_tx_wait", default=5, help="The TNC will wait this value in seconds", type=int)
+    group.add_argument('--max-tx-wait', dest="max_tx_wait", default=7, help="The TNC will wait this value in seconds", type=int)
+    group.add_argument('--max-packets-tx', dest="max_packets", default=1, help="The number of frames that can be TXed at once. Set to -1 to send all", type=int)
 
     args = parser.parse_args()
 
@@ -39,19 +44,18 @@ def main():
         sys.exit(0)
 
 
-    tx_inhibit = Lock()
 
     def kiss_rx_callback(frame: bytes):
         logging.debug(f"Received KISS frame: {frame.hex()}")
-        tx_inhibit.acquire()
         if args.tx:
-            radio.tx([frame]) #TODO : this is where we could send multiple packets but a lot of protocols except one at a time
-        tx_inhibit.release()
+            radio.tx([frame])
 
     def rf_rx_callback(packet: bytes):
         logging.debug(f"Received RF packet: {packet.hex()}")
         if tnc_interface:
             tnc_interface.tx(packet)
+        if tcp_interface:
+            tcp_interface.tx(packet)
         if args.stdout:
             sys.stdout.buffer.write(packet)
             sys.stdout.flush()
@@ -73,7 +77,7 @@ def main():
     try:
         modem = freedv.FreeDV()
     except OSError:
-        logger.error("Could not find libcodec2.so - please ensure it's installed and ldconfig has been run")
+        logger.error("Could not find libcodec2 - please ensure it's installed and ldconfig has been run")
         sys.exit(0)
 
     if sys.platform == "darwin":
@@ -87,6 +91,11 @@ def main():
     else:
         tnc_interface = None
 
+    if args.tcp:
+        tcp_interface = tnc.KissTCPInterface(kiss_rx_callback)
+    else:
+        tcp_interface = None
+
     if args.tx == False:
         args.tx_sound_device = None 
     try:
@@ -98,14 +107,21 @@ def main():
                         callback=rf_rx_callback,
                         rig=rig,
                         preamble_frame_count=args.preamble_length,
-                        post_tx_wait=args.min_tx_wait
+                        post_tx_wait_min=args.min_tx_wait,
+                        post_tx_wait_max=args.max_tx_wait,
+                        max_packets=args.max_packets
                     )
     except UnboundLocalError:
        logger.error("Couldn't intialize RF. Likely your soundcard isn't avaliable")
        sys.exit()
-
+    
     while True:
         radio.rx()
 
+
+
+            
 if __name__ == "__main__":
-    main()
\ No newline at end of file
+    main()
+
+
